/*  Csum.c by B.Mours --- Oct 30, 2000    
    Channel Summary: Compute chisquare for all channels                      */

#include "Csum.h"
#include "Frv.h"
#include <string.h>
#include <math.h>

void CsumResProc(FrameH *frame,FrvRes *res,FrVect *vect); /*-private function*/

/*--------------------------------------------------------------- CsumInit--*/
int CsumInit(Csum *cs, FrameH *frame)
/*---------------------------------------------------------------------------*/
{FrAdcData *adc;
 int nAdc, i;

 if(frame->rawData == NULL) return(0);

 nAdc = 0;
 for(adc = frame->rawData->firstAdc; adc != NULL; adc=adc->next)
   {if(adc->data->nData < cs->nDataMin) continue;
    if(adc->data->nData > cs->nDataMax) continue;
    if(strstr(adc->name,"GPS") != NULL) continue;  /*--- skip the GPS channels--*/ 
    nAdc ++;;}

 printf(">>CsumInit called for %d ADC channels\n",nAdc);

 if(nAdc == 0) return(NULL);

 if((cs->chi2     = FrVectNew1D("algorithms",FR_VECT_STRING, nAdc,1.,
                          "channel id","pointer"       )) == NULL) return(0);
 if((cs->index    = FrVectNew1D("index",FR_VECT_4S, nAdc,1.,
                          "channel id","index"         )) == NULL) return(0);
 if((cs->previous = FrVectNew1D("prev",FR_VECT_8R, nAdc,1.,
                          "channel id","previous value")) == NULL) return(0);
   
 for(i=0; i<nAdc; i++)  {cs->index->dataI[i] = i;} /*-- default index table--*/
   
 cs->GTimeS = frame->GTimeS;

           /*------- build the algorithms to compute the residuals ---------*/

 cs->meanRes = FrvResNew(cs->resPeriod, cs->resScale);
 cs->rmsRes  = FrvResNew(cs->resPeriod, cs->resScale);
 cs->pwr1Res = FrvResNew(cs->resPeriod, cs->resScale);
 cs->pwr2Res = FrvResNew(cs->resPeriod, cs->resScale);
 cs->pwr3Res = FrvResNew(cs->resPeriod, cs->resScale);
 cs->deltaRes= FrvResNew(cs->resPeriod, cs->resScale);
 cs->slowRes = FrvResNew(cs->resPeriod, cs->resScale);

 return(1);
}
/*--------------------------------------------------------------- CsumNew --*/
Csum* CsumNew(int nDataMin, 
              int nDataMax, 
              double threshold,
              char * prefix)
/*---------------------------------------------------------------------------*/
{Csum *cs;

 cs = (Csum *) calloc(1,sizeof(Csum));
 if(cs == NULL) return(NULL);

 cs->nDataMin  = nDataMin;
 cs->nDataMax  = nDataMax;
 cs->threshold = threshold;
 cs->nDataSlow = 16;
 cs->resPeriod = 50;
 cs->resScale = 100.;

 if(FrStrCpy(&(cs->prefix),prefix) == NULL) return(NULL);

 cs->nCall = 0;
                
 return(cs);
}

/*--------------------------------------------------------------- CsumProc --*/
int CsumProc(Csum *cs, 
             FrameH *frame)
/*---------------------------------------------------------------------------*/
{FrvChi2 *chi2;
 FrAdcData *adc;
 FrVect *names, *rates, *chi2Vect, *meanVect, *rmsVect, 
        *pwr1Vect, *pwr2Vect, *pwr3Vect, *deltaVect;
 int iAdc, nAdc, index, i, nprint, nData, outSize;
 double meanChi2, mean, rms, chi2v;
 char buf[24], nameC[32], nameN[32], nameM[32], nameR[32], nameRms[32], 
       nameP1[32], nameP2[32], nameP3[32], nameD[32], **namesPtr;
 
 if(frame == NULL)          return(0);
 if(frame->rawData == NULL) return(0);

 sprintf(nameC,"%s:CsFChi2" ,cs->prefix);
 sprintf(nameN,"%s:CsFNames",cs->prefix);
 sprintf(nameM,"%s:CsFMean" ,cs->prefix);
 sprintf(nameR,"%s:CsFRates",cs->prefix);
 sprintf(nameRms,"%s:CsFRms",cs->prefix);
 sprintf(nameP1,"%s:CsFPwr32-128",cs->prefix);
 sprintf(nameP2,"%s:CsFPwr128-1k",cs->prefix);
 sprintf(nameP3,"%s:CsFPwr1k-8k",cs->prefix);
 sprintf(nameD,"%s:CsFDelta",cs->prefix);


                 /*-----------------------  Init algorithms if needed -------*/

 if(cs->nCall == 0)
   {if(CsumInit(cs, frame) == 0) return(0);}

                /*------ create summary structures -------*/

 nAdc = cs->chi2->nData;
 names     = FrVectNew1D(nameN  ,FR_VECT_STRING, nAdc,1.,"channel id","name");
 rates     = FrVectNew1D(nameR  ,FR_VECT_2U, nAdc,1.,"channel id","rates");
 chi2Vect  = FrVectNew1D(nameC  ,FR_VECT_2U, nAdc,1.,"channel id","chi2(%)");
 meanVect  = FrVectNew1D(nameM  ,FR_VECT_4R, nAdc,1.,"channel id","mean");
 rmsVect   = FrVectNew1D(nameRms,FR_VECT_4R, nAdc,1.,"channel id","rms(%)");
 pwr1Vect  = FrVectNew1D(nameP1 ,FR_VECT_4R, nAdc,1.,"channel id","power(%)");
 pwr2Vect  = FrVectNew1D(nameP2 ,FR_VECT_4R, nAdc,1.,"channel id","power(%)");
 pwr3Vect  = FrVectNew1D(nameP3 ,FR_VECT_4R, nAdc,1.,"channel id","power(%)");
 deltaVect = FrVectNew1D(nameD  ,FR_VECT_8R, nAdc,1.,"channel id","delta");
 if((names    == NULL) ||
    (rates    == NULL) ||
    (meanVect == NULL) ||
    (chi2Vect == NULL) ||
    (rmsVect  == NULL) ||
    (pwr1Vect == NULL) ||
    (pwr2Vect == NULL) ||
    (pwr3Vect == NULL) ||
    (deltaVect== NULL)) return(0);

               /*------------ loop over all chanels-----*/

 meanChi2 = 0.;
 iAdc =-1;
 namesPtr = names->dataQ;
 nprint = 0;

 for(adc = frame->rawData->firstAdc; adc != NULL; adc=adc->next)
   {if(adc->data->nData < cs->nDataMin) continue;
    if(adc->data->nData > cs->nDataMax) continue;
    if(strstr(adc->name,"GPS") != NULL) continue;  /*--- skip the GPS channels--*/ 
    iAdc++;
    if(iAdc >= cs->index->nData) break;

                /*--------------- fix adc->vector names if needed -----------*/

    if(strcmp(adc->name,adc->data->name) != 0)
       {free(adc->data->name);
        FrStrCpy(&adc->data->name, adc->name);}

    if(cs->nCall > 0)
       {index = cs->index->dataI[iAdc];
        if(index == cs->index->nData)  continue;         
        chi2 = (FrvChi2*) cs->chi2->dataQ[index];

           /*---- check that we get the proper object -----*/

        if(strcmp(chi2->name, adc->name) != 0)
          {if(nprint++ < 10)
             printf("CsumProc: adc order has changed(%s/%s) %d index=%d\n",
                           adc->name, chi2->name,iAdc, index);
           for(i=0; i<cs->index->nData; i++)
              {chi2 = (FrvChi2*) cs->chi2->dataQ[i];
               if(strcmp(chi2->name, adc->name) == 0) break;;}
             
           cs->index->dataI[iAdc] = i;
           if(i == cs->index->nData) 
	      {printf(" this is a new channel:%s\n",adc->name);
               continue;}}}
    else{index = iAdc;
         chi2 = NULL;}

               /*--------------------------- compute delta ------------------*/
   
    if(cs->nCall == 0) 
        {cs->previous->dataD[iAdc] = FrVectGetV(adc->data, 0);}

    else{FrvDelta(adc->data,&(deltaVect->dataD[index]),&(cs->previous->dataD[index]));}

               /*---------------------- decimate fast channels--------------*/
   
    if(adc->data->nData > 5000) FrvDecimate(adc->data, 2);
    
             /*----- Get the processing algorithm and init if needed ------*/

    if(cs->nCall == 0)
       {nData = adc->data->nData;
        if     (nData < 500)    {outSize = nData;}
        else if(nData%256 == 0) {outSize = 512;}
        else                    {outSize = 250;}
 
        chi2 = FrvChi2New(outSize, 1, cs->threshold);
        FrvChi2Ini(chi2, adc->data);
        cs->chi2->dataQ[iAdc] = (char *) chi2;}

              /*--- compute first the mean and rms -------*/

    FrvRms(adc->data, &mean, &rms); 
    meanVect->dataF[index] = mean; 
    rmsVect ->dataF[index] = rms;

    if(rms == 0.)
      {chi2Vect->dataUS[index] = 0.;
       pwr1Vect->dataF[index] = 0.;
       pwr2Vect->dataF[index] = 0.;
       pwr3Vect->dataF[index] = 0.;}

    else
      {FrvChi2Proc(chi2, adc->data);

            /*------- store the chi2 result ----*/

       if(chi2->chi2 < 1.) 
            {chi2v = 100.*chi2->chi2;}
       else {chi2v = 100.*(1.+log(chi2->chi2));}

       chi2Vect->dataUS[index] = chi2v;
       meanChi2 += chi2v;

       /*--- compute the mean power and store it ----*/

       pwr1Vect->dataF[index] = chi2->power1;
       pwr2Vect->dataF[index] = chi2->power2;
       pwr3Vect->dataF[index] = chi2->power3;}

      /*--- static data -----*/

    rates->dataUS[index] = adc->data->nData;

    FrStrCpy(&(namesPtr[index]), adc->name);

    }

 chi2Vect->GTimeS = frame->GTimeS;
 chi2Vect->GTimeN = frame->GTimeN;

 cs->nCall++;
 cs->meanChi2 = meanChi2/cs->chi2->nData;

                   /*----- fill chi2square summary info ---*/

 sprintf(buf,"<chi2>=%6.2f",cs->meanChi2);
 
 FrSummaryNew(frame, nameC,  "Names in XCsFName", buf, chi2Vect, NULL);

        /*------------ create static data for the channel description ---*/

 if(frame->detectProc == NULL)
    frame->detectProc = FrDetectorNew("detectProc");
 if(frame->detectProc == NULL) return(0);

 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(nameR, "Fast channel rates",NULL,cs->GTimeS,0,0,rates,NULL));
 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(nameN, "Fast chanel names",NULL, cs->GTimeS,0,0,names,NULL));

                  /*------ compute residuals ------------------*/

 CsumResProc(frame, cs->meanRes, meanVect);
 CsumResProc(frame, cs->rmsRes,  rmsVect);
 CsumResProc(frame, cs->pwr1Res, pwr1Vect);
 CsumResProc(frame, cs->pwr2Res, pwr2Vect);
 CsumResProc(frame, cs->pwr3Res, pwr3Vect);

 CsumResProc(frame, cs->deltaRes, deltaVect);

       /*------------------- fill slow channel info ---*/

 CsumSlow(cs, frame);

 return(1);

}
/*----------------------------------------------------------- CsumResStore --*/
void CsumResProc(FrameH *frame,
                 FrvRes *res,
                 FrVect *vect)
/*---------------------------------------------------------------------------*/
/* Compute and store residual and associated information                     */
{FrVect *copyS, *copyF;
 char buf[128];

 FrvResProc(res, vect);

 copyS = FrvCopyToS(res->residual,NULL);
 FrSummaryNew(frame, vect->name,   "Names in XCsFName", NULL, copyS, NULL);

 sprintf(buf,"%s_<>",res->name);
 copyF = FrvCopyToF(res->mean, NULL);
 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(buf, NULL,NULL, copyF->GTimeS,0,0, copyF,NULL));

 sprintf(buf,"%s_rms",res->name);
 copyF = FrvCopyToF(res->oneOverRms, NULL);
 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(buf, NULL,NULL, copyF->GTimeS,0,0, copyF,NULL));

 FrVectFree(vect);

 return;
}
/*-------------------------------------------------------------- CsumSlow --*/
int CsumSlow(Csum *cs, FrameH *frame)
/*--------------------------------------------------------------------------*/
/* buid summary information for slow adc channels                           */
/*--------------------------------------------------------------------------*/
{FrAdcData *adc;
 FrVect *names, *meanVect;
 int nAdc, iAdc;
 static int first = 0;
 double mean;
 char nameN[32], nameM[32], **namesPtr;
 
 if(frame->rawData == NULL) return(0);

 sprintf(nameN,"%s:CsSNames",cs->prefix);
 sprintf(nameM,"%s:CsSMean" ,cs->prefix);
 
 nAdc = 0;
 for(adc = frame->rawData->firstAdc; adc != NULL; adc=adc->next)
   {if(adc->data->nData <= cs->nDataSlow) nAdc++;}

 if(first == 0)
      {first = 1;
      printf(">>CsumSlow called for %d ADC channels\n",nAdc);}

 if(nAdc == 0) return(NULL);
    
                /*------ create summary structures -------*/

 names     = FrVectNew1D(nameN, FR_VECT_STRING, nAdc,1.,"channel id","name");
 meanVect  = FrVectNew1D(nameM, FR_VECT_4R,     nAdc,1.,"channel id","mean");
 if((names    == NULL) ||
    (meanVect == NULL)) return(0);

               /*------------ loop over all chanels-----*/

 iAdc = 0;
 namesPtr = names->dataQ;

 for(adc = frame->rawData->firstAdc; adc != NULL; adc=adc->next)
   {if(adc->data->nData > cs->nDataSlow) continue;
   
    FrvMean(adc->data, &mean); 
    meanVect->dataF[iAdc] = mean;
    FrStrCpy(&(namesPtr[iAdc]), adc->name);

    iAdc++;}

                   /*----- fill summary info ---*/

 CsumResProc(frame, cs->slowRes, meanVect);
 
       /*------------ create static data for the channel description ---*/

 if(frame->detectProc == NULL)
    frame->detectProc = FrDetectorNew("detectProc");
 if(frame->detectProc == NULL) return(0);

 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(nameN, "Slow channel names","", cs->GTimeS,0,0,names, NULL));


 return(1);
}
/*-------------------------------------------------------------- CsumSaveS --*/
int CsumSaveS(Csum *cs, 
              FrameH *frame)
/*---------------------------------------------------------------------------*/
{FrvChi2 *chi2;
 FrVect *root, *spectrum;
 int i;
 char name[32];

 if(cs    == NULL) return(0);
 if(frame == NULL) return(0);

               /*----------------- Copy all the spectrum ------------------*/

 root = NULL;
 for(i = 0; i<cs->index->nData; i++)
   {chi2 = (FrvChi2*) cs->chi2->dataQ[i];
    if(chi2 == NULL) return(0);

    spectrum = FrvCopyToF(chi2->fft->amplitudeA, chi2->name);
    if(spectrum == NULL) return(0);

    spectrum->next = root;
    root = spectrum;}

             /*---------------- Create static data to hold the spectrum ---*/

 if(frame->detectProc == NULL)
    frame->detectProc = FrDetectorNew("detectProc");
 if(frame->detectProc == NULL) return(0);

 sprintf(name,"%s:CsFSpectrum",cs->prefix);
 FrStatDataAdd(frame->detectProc,
    FrStatDataNew(name,   "Fast Channel Spectrum",
                          NULL,frame->GTimeS,0,0,root,NULL));

 return(1);

}
/*-------------------------------------------------------------- CsumDumpF--*/
void CsumDumpF(FrameH *frame, char* prefix, char* tag, FILE* out)
/*--------------------------------------------------------------------------*/
/* Produce a formated dump of the Channel Statistic                         */
/*--------------------------------------------------------------------------*/
{FrVect *names, *means, *meansM, *meansR, 
        *chi2, *rates, *rms, *rmsM, *rmsR, *pwr1, *pwr2, *pwr3;
 int i;
 char buf[32];

 if(frame == NULL) return;
 if(prefix == NULL) return;

 fprintf(out,"\n Dump Fast Channel Summary for detector %s",prefix);
 fprintf(out,"  GPS=%d\n",frame->GTimeS);

 sprintf(buf,"%s:CsFNames",prefix);
 names = FrStatDataGetV(frame, buf);
 if(names == NULL)
   {fprintf(out," No name list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFRates",prefix);
 rates = FrStatDataGetV(frame, buf);
 if(rates == NULL)
   {fprintf(out," No name list found (%s)\n",buf);
    return;}
             /*------------------ Mean Values ---------*/

 sprintf(buf,"%s:CsFMean",prefix);
 means = FrSummaryGetV(frame, buf);
 if(means == NULL)
   {fprintf(out," No mean list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFMean_<>",prefix);
 meansM = FrStatDataGetV(frame, buf);
 if(meansM == NULL)
   {fprintf(out," No mean mean list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFMean_rms",prefix);
 meansR = FrStatDataGetV(frame, buf);
 if(meansR == NULL)
   {fprintf(out," No mean rms list found (%s)\n",buf);
    return;}
 
           /*---------- rmsValues ---------------*/

 sprintf(buf,"%s:CsFRms",prefix);
 rms = FrSummaryGetV(frame, buf);
 if(rms == NULL)
   {fprintf(out," No mean list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFRms_<>",prefix);
 rmsM = FrStatDataGetV(frame, buf);
 if(rmsM == NULL)
   {fprintf(out," No rms mean list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFRms_rms",prefix);
 rmsR = FrStatDataGetV(frame, buf);
 if(rmsR == NULL)
   {fprintf(out," No rms rms list found (%s)\n",buf);
    return;}
 

 sprintf(buf,"%s:CsFChi2",prefix);
 chi2 = FrSummaryGetV(frame, buf);
 if(chi2 == NULL)
   {fprintf(out," No rms list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFPwr32-128",prefix);
 pwr1 = FrSummaryGetV(frame, buf);
 if(pwr1 == NULL)
   {fprintf(out," No rms list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFPwr128-1k",prefix);
 pwr2 = FrSummaryGetV(frame, buf);
 if(pwr2 == NULL)
   {fprintf(out," No rms list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsFPwr1k-8k",prefix);
 pwr3 = FrSummaryGetV(frame, buf);
 if(pwr3 == NULL)
   {fprintf(out," No rms list found (%s)\n",buf);
    return;}

 fprintf(out,"   #     Names                      rates "
             "  ---------- mean -------"
             " --------- rms -------- "
             "chi2 Pwr32-128-1k-4k\n");
 for(i=0; i<names->nData; i++)
   {fprintf(out,"%4d %-30s%6d",i,names->dataQ[i],rates->dataUS[i]);
    fprintf(out,"%5d(%10.2f%9.2f)",means->dataS[i],
                          meansM->dataF[i],1./meansR->dataF[i]);
    fprintf(out,"%5d(%9.2f%8.2f)",rms->dataS[i],
                          rmsM->dataF[i],1./rmsR->dataF[i]);
    fprintf(out,"%5d",chi2->dataS[i]);
    fprintf(out,"%5d%5d%5d\n",pwr1->dataS[i],pwr2->dataS[i],pwr3->dataS[i]);}

 return;}
/*-------------------------------------------------------------- CsumDumpS--*/
void CsumDumpS(FrameH *frame, char* prefix, char* tag, FILE* out)
/*--------------------------------------------------------------------------*/
/* Produce a formated dump of the Channel Statistic                         */
/*--------------------------------------------------------------------------*/
{FrVect *names, *means;
 int i;
 char buf[32];
 float val;

 if(frame == NULL) return;
 if(prefix == NULL) return;

 fprintf(out,"\n Dump Slow Channel Summary for detector %s",prefix);
 fprintf(out,"  GPS=%d\n",frame->GTimeS);

 sprintf(buf,"%s:CsSNames",prefix);
 names = FrStatDataGetV(frame, buf);
 if(names == NULL)
   {fprintf(out," No name list found (%s)\n",buf);
    return;}

 sprintf(buf,"%s:CsSMean",prefix);
 means = FrSummaryGetV(frame, buf);
 if(means == NULL)
   {fprintf(out," No mean list found (%s)\n",buf);
    return;}

 fprintf(out,"   #     Names                         mean values\n");
 for(i=0; i<names->nData; i++)
   {fprintf(out,"%4d %-30s",i,names->dataQ[i]);
    val = means->dataF[i];
    if(val < 0) val = -val;
    if((val > 1.e-3) && (val < 1.))
         {fprintf(out," %11.5f\n",means->dataF[i]);}
    else if((val >=1.)  && (val < 100.))
         {fprintf(out," %10.4f\n",means->dataF[i]);}
    else if((val >=100.)  && (val < 1.e+4))
         {fprintf(out," %8.2f\n",means->dataF[i]);}
    else if(val == 0.)
         {fprintf(out,"     0.\n");}
    else {fprintf(out," %e\n",means->dataF[i]);}}

 return;}
