/*  Author : B.Mours and D.Buskulic LAPP (Annecy) Sept 13, 2000           */

/*//////////////////////////////////////////////////////////////////// */
/*                                                                     */
/*   Fast Fourier Transform                                            */
/*   Uses the FFTW package if USE_FFTW is defined                      */
/*   If not, uses a simpler fft algorithm included                     */
/*                                                                     */
/*//////////////////////////////////////////////////////////////////// */

#include "FrvFFT.h"
#include <string.h>
#include <math.h>

char FRVERROR[258];

/*------------------------------------------------------------ FrvRFFTdoD --*/
void FrvRFFTdoD(FrvRFFT *fft, FrVect *vect)
/*---------------------------------------------------------------------------*/
{int i, nData;
 double *input, *temp,  mean, *window, *output;


         /*---------------------------------------------- prepare data ------*/
         /* remark: the output buffer is used to prepare the input data ...*/

 nData  = fft->nData;
 input  = fft->output->dataD;  
 temp   = fft->output->dataD + nData;  

         /*---------------------------------------- extract input data-------*/

 if(vect->type == FR_VECT_2S)
       {for(i=0; i<nData; i++) {input[i] = vect->dataS[i];}}
 else if(vect->type == FR_VECT_4R)
       {for(i=0; i<nData; i++) {input[i] = vect->dataF[i];}}
 else if(vect->type == FR_VECT_4S)
       {for(i=0; i<nData; i++) {input[i] = vect->dataI[i];}}
 else if(vect->type == FR_VECT_8R)
       {for(i=0; i<nData; i++) {input[i] = vect->dataD[i];}}
     
         /*----------------- compute and supress the mean value if needed---*/

 mean = 0.; 
 if((fft->option & 2) == 2)
   {for(i=0; i<nData; i++) {mean += input[i];}
    mean = mean/nData;
    fft->mean = mean;}

         /*------------------------------- prepare input data with window ---*/

 if     ((fft->option & 3) == 2)
   {for(i=0; i<nData; i++) {input[i] = input[i] - mean;}}
 else if((fft->option & 3) == 1)
   {window = fft->window->dataD;
     for(i=0; i<nData; i++) {input[i] = window[i]*input[i];}}
 else if((fft->option & 3) == 3)
   {window = fft->window->dataD;
    for(i=0; i<nData; i++) {input[i] = window[i]*(input[i] - mean);}}

         /*---------------------------------------------------- call fft ---*/

 rfftw_one(fft->plan, (fftw_real*) input, (fftw_real*) temp);

         /* --------------- copy the result to the FrVect complex order ----*/

 output = fft->output->dataD;
 for (i=0;i<((nData+1)/2)-1;i++) 
      {output[2*i]   = temp[i+1];
       output[2*i+1] = temp[nData-i];}
      
 fft->mean = temp[0];    /*--- DC component ---*/

 if (nData % 2 ==0) 
      {output[nData-2] = temp[nData/2];   /* Nyquist frequency */
       output[nData-1] = 0.;} 
 
 return;
}
/*------------------------------------------------------------ FrvRFFTdoF --*/
void FrvRFFTdoF(FrvRFFT *fft, FrVect *vect)
/*---------------------------------------------------------------------------*/
{int i, nData;
 float  *input, *temp,  mean;
 double *window, *output;


         /*---------------------------------------------- prepare data ------*/
         /* remark: the output buffer is used to prepare the input data ...*/

 nData  = fft->nData;
 input  = fft->output->dataF;  
 temp   = fft->output->dataF + nData;  

         /*---------------------------------------- extract input data-------*/

 if(vect->type == FR_VECT_2S)
       {for(i=0; i<nData; i++) {input[i] = vect->dataS[i];}}
 else if(vect->type == FR_VECT_4R)
       {for(i=0; i<nData; i++) {input[i] = vect->dataF[i];}}
 else if(vect->type == FR_VECT_4S)
       {for(i=0; i<nData; i++) {input[i] = vect->dataI[i];}}
 else if(vect->type == FR_VECT_8R)
       {for(i=0; i<nData; i++) {input[i] = vect->dataD[i];}}
     
         /*----------------- compute and supress the mean value if needed---*/

 mean = 0.; 
 if((fft->option & 2) == 2)
   {for(i=0; i<nData; i++) {mean += input[i];}
    mean = mean/nData;
    fft->mean = mean;}

         /*------------------------------- prepare input data with window ---*/

 if     ((fft->option & 3) == 2)
   {for(i=0; i<nData; i++) {input[i] = input[i] - mean;}}
 else if((fft->option & 3) == 1)
   {window = fft->window->dataD;
     for(i=0; i<nData; i++) {input[i] = window[i]*input[i];}}
 else if((fft->option & 3) == 3)
   {window = fft->window->dataD;
    for(i=0; i<nData; i++) {input[i] = window[i]*(input[i] - mean);}}

         /*---------------------------------------------------- call fft ---*/

 rfftw_one(fft->plan, (fftw_real*) input, (fftw_real*) temp);

         /* --------------- copy the result to the FrVect complex order ----*/

 output = fft->output->dataD;
 for (i=0;i<((nData+1)/2)-1;i++) 
      {output[2*i]   = temp[i+1];
       output[2*i+1] = temp[nData-i];}
      
 fft->mean = temp[0];    /*--- DC component ---*/

 if (nData % 2 ==0) 
      {output[nData-2] = temp[nData/2];   /* Nyquist frequency */
       output[nData-1] = 0.;} 
 

 return;

}
/*------------------------------------------------------------ FrvRFFTFree --*/
void  FrvRFFTFree(FrvRFFT* fft)
/*---------------------------------------------------------------------------*/
{
  FrVectFree(fft->output);

  if(fft->window != NULL) FrVectFree(fft->window);
  if(fft->amplitude  != NULL) FrVectFree(fft->amplitude);
  if(fft->amplitudeA != NULL) FrVectFree(fft->amplitudeA);

  fftw_destroy_plan(fft->plan);

  free(fft);

  return;
}
/*------------------------------------------------------------ FrvRFFTFor --*/
FrvRFFT* FrvRFFTFor(FrvRFFT *fft, FrVect *vect)
/*---------------------------------------------------------------------------*/
/*   Applies the FFT algorithm "fft" to the vector "vect"                    */
/*   Returns the result in the FrvRFFT structure as output                   */
/*   If on input fft=0, builds a standard fft structure from the vector vect */
/*  with the option "option" (remark: The option is used only if fft = NULL) */
/*---------------------------------------------------------------------------*/
{
 int i, nData;
 double *amplitude, *amplitudeA, *output, decay;

 if(vect == NULL)  return(NULL);

          /*--  If no fft algorithm object on input, build a standard one ---*/

 if(fft  == NULL)  
   {fft = FrvRFFTNew(0);
   if(fft == NULL) return(NULL);}

 if(fft->output == NULL)
   {if(FrvRFFTInit(fft, vect) == 0) return(NULL);}

          /*----------------------- check that we work on the same buffer ---*/

 if(fft->nData != vect->nData) 
   {sprintf(FRVERROR," nData missmatch:%ld %d\n",fft->nData,vect->nData);
    return(NULL);}

         /*-------------------- set start time for all the series defined ---*/

 fft->output->GTimeS  = vect->GTimeS;
 fft->output->GTimeN  = vect->GTimeN;
 if(fft->amplitude != NULL)
   {fft->amplitude->GTimeS  = vect->GTimeS;
    fft->amplitude->GTimeN  = vect->GTimeN;}
 if(fft->amplitudeA != NULL)
   {fft->amplitudeA->GTimeS = vect->GTimeS;
    fft->amplitudeA->GTimeN = vect->GTimeN;}

         /*-----------compute the FFT acroding the FFT library version ------*/

 nData  = fft->nData;
 if(fftw_sizeof_fftw_real() == 4) 
      FrvRFFTdoF(fft, vect);
 else FrvRFFTdoD(fft, vect);
 
         /*--------------------------------------------- compute amplitude---*/

 output = fft->output->dataD;
 if((fft->option & 4) == 4)
   {amplitude = fft->amplitude->dataD;
    for(i=0; i<nData/2; i++)
      {amplitude[i] = sqrt(output[2*i]  *output[2*i]+
                           output[2*i+1]*output[2*i+1]);}}

        /*------------------------------------- compute average amplitude ---*/

 if((fft->option & 8) == 8)
   {amplitudeA = fft->amplitudeA->dataD;
    amplitude  = fft->amplitude->dataD;
    
    fft->nDecay++;
    if(fft->nDecay == 1)
      {for(i=0; i<nData/2; i++)
         {amplitudeA[i] =  amplitude[i];}}
    else
      {decay = 1.-1./fft->nDecay;
       if(fft->decay < decay) decay = fft->decay;
       for(i=0; i<nData/2; i++)
         {amplitudeA[i] =  decay*amplitudeA[i] + 
                       (1.-decay)*amplitude[i];}}}

 fft->nFFT++;

 return(fft);

}

/*-------------------------------------------------------- FrvRFFTGetError --*/
char*  FrvRFFTGetError()
/*---------------------------------------------------------------------------*/
{
  return(FRVERROR);
}

/*------------------------------------------------------------ FrvRFFTInit --*/
int FrvRFFTInit(FrvRFFT* fft, FrVect *vect)
/*---------------------------------------------------------------------------*/
/*   Initilized the fft object: creat all output structures                  */
/*   Returns: 0 in case of probleme, 1 otherwise                             */
/*---------------------------------------------------------------------------*/
{int i;
 double a, b;

 if(vect == NULL) 
    {sprintf(FRVERROR,"no input vector");
     return(NULL);}

 if((vect->type != FR_VECT_2S) &&
    (vect->type != FR_VECT_4R) &&
    (vect->type != FR_VECT_4S) &&
    (vect->type != FR_VECT_8R)) 
    {sprintf(FRVERROR,"input type not supported:%d",vect->type);
     return(NULL);}

 fft->nData  = vect->nData;
 fft->dt     = vect->dx[0];
 fft->dnu    = 1./(fft->nData*fft->dt);
 fft->decay = .99;
 fft->nDecay = 0;
    
          /*---- init for output space ----*/

 fft->output  = FrVectNew1D(vect->name, FR_VECT_C16,fft->nData, 
                            fft->dnu, "frequency[Hz]","fft");
 if(fft->output  == NULL) 
    {sprintf(FRVERROR,"malloc in/out failed:ndata=%ld",fft->nData);
     return(NULL);}

          /*---- only half of the output is seen by the user ----*/

 fft->output->nData = fft->nData/2;   
    
          /*------------------------- init for the window ----*/
          /*----- the normalization do not change the spectrum amplitude ---*/

 if((fft->option & 1) == 1)
   {fft->window  = FrVectNew1D("window",FR_VECT_8R, fft->nData,
                                fft->dt, "time","window");
    if(fft->window  == NULL) 
       {sprintf(FRVERROR,"malloc window failed:ndata=%ld",fft->nData);
        return(NULL);}

    a = fft->nData - 1.;
    b = 6.2831853071795864769/a;

    for (i=0; i<fft->nData; i++)
      {fft->window->dataD[i] = 1. + cos(b*((double)i - a*.5));}}

        /*------ init for the average amplitude ----*/

 if((fft->option & 8) == 8)
   {fft->amplitudeA  = FrVectNew1D(vect->name,  FR_VECT_8R, fft->nData/2,
                                   fft->dnu,"frequency[Hz]","<amplitude>");
    if( fft->amplitudeA  == NULL)  
       {sprintf(FRVERROR,"malloc amplitudeA failed:ndata=%ld",fft->nData);
        return(NULL);}
 
    for (i=0; i<fft->nData/2; i++)
       {fft->amplitudeA->dataD[i] = 0.;}}
                
        /*------ init for the amplitude spectrum----*/

 if((fft->option & 4) == 4)
   {fft->amplitude  = FrVectNew1D(vect->name, FR_VECT_8R, fft->nData/2,
                                  fft->dnu,"frequency[Hz]","amplitude");
    if( fft->amplitude  == NULL)  
       {sprintf(FRVERROR,"malloc amplitude failed:ndata=%ld",fft->nData);
        return(NULL);}}
      
         /*-------------- specific FFT init ---*/
 
      /*--- creat FFTW plan  ---*/
 
 fft->plan = rfftw_create_plan(vect->nData,FFTW_REAL_TO_COMPLEX,FFTW_ESTIMATE);
 if(fft->plan == NULL) return(0);

      /*---- adjust bin origin ----------*/

 fft->output->startX[0] = 0.5*fft->dnu;
 if(fft->amplitude  != NULL)  fft->amplitude->startX[0] = 0.5*fft->dnu;
 if(fft->amplitudeA != NULL) fft->amplitudeA->startX[0] = 0.5*fft->dnu;

 return(1);

}
/*------------------------------------------------------------- FrvRFFTNew --*/
FrvRFFT* FrvRFFTNew(int option)
/*---------------------------------------------------------------------------*/
/*   Returns the FrvRFFT structure as output            

  option: 
   bit 0  hanning window (yes = 1) (normalized to not change the amplitude)
   bit 1  average suppress (yes = 1)
   bit 2  amplitude spectrum computed (yes = 1)
   bit 3  average amplitude spectrum computed (yes = 1)

   option = 5 means hanning window+amplitude spectrum computed               */
/*---------------------------------------------------------------------------*/
{
 FrvRFFT *fft;

      /*-------- create the fft structure ----*/

 fft = (FrvRFFT *) (calloc(1,sizeof(FrvRFFT)));
 if(fft == NULL) 
   {sprintf(FRVERROR,"malloc fft failed");
    return(NULL);}
 
 fft->option = option;

         /*-- we need to compute the amplitude spectrum to get the mean */

 if((fft->option & 8) == 8)
    {fft->option = (fft->option | 4);}

 sprintf(FRVERROR,"no error");

 return(fft);

}

/*------------------------------------------------------------- FrvFFTD704 --*/
void FrvFFTD704(double *a, long msign)
/*---------------------------------------------------------------------------*/
/* cdfft computes the finite Fourier transform of a complex periodic         */
/* sequence, whose period n must be a power of 2 (n=2**|msign|).             */
/* To ensure optimum use of storage, the same array is used for input        */
/* and output. If msign is positive the direct Fourier transform is          */
/* computed. If msign is negative, the inverse Fourier transform is          */
/* computed. This code is a c translation of the routine D704 (CFFT)         */
/* from CERNLIB. See the CERN writeup for more information.                  */
/* Author: B. Mours        27/02/91                                          */
/*---------------------------------------------------------------------------*/
 
{
 double c,s,ur,ui,ur0,wr,wi,tr,ti,temp;
 long  m,n,nv2,nm1,i,j,k,le,l,le1,ip;
 long indexI,indexJ,indexIp;
 
 if (msign == 0) return;
 m   = msign;
 if (msign < 0) m = -m;
 n   = 1;
 for (i=0; i<m; i++) {n=2*n;}
 nv2 = n/2;
 nm1 = n-1;
 j   = 0;
 
 for (i=0; i<nm1; i++)
    {indexI = 2*i;
     if (i < j)
        {indexJ      = 2*j;
         tr          = a[indexJ  ];
         ti          = a[indexJ+1];
         a[indexJ  ] = a[indexI  ];
         a[indexJ+1] = a[indexI+1];
         a[indexI  ] = tr;
         a[indexI+1] = ti;
        }
     k = nv2-1;
     while ( k < j )
        {j = j - k - 1;
         k = (k-1) / 2;
        }
    j = j + k + 1;
   }
 
 for (i=0; i<n; i=i+2)
    {indexI       = 2*i;
     tr           = a[2*(i+1)];
     ti           = a[2*(i+1)+1];
     a[2*(i+1)  ] = a[indexI  ] - tr;
     a[2*(i+1)+1] = a[indexI+1] - ti;
     a[indexI  ]  = a[indexI  ] + tr;
     a[indexI+1]  = a[indexI+1] + ti;
    }
 
 if (m == 1) return;
 c = 0;
 s = 1.;
 if (msign < 0) s = -1.;
 le = 2;
 
 for (l=2; l<=m; l++)
    {wr  = c;
     wi  = s;
     ur  = wr;
     ui  = wi;
     c   = sqrt(c*.5+.5);
     s   = wi/(c+c);
     le1 = le;
     le  = le1+le1;
 
     for (i=0; i<n; i=i+le)
        {ip      = i + le1;
         indexI  = 2*i;
         indexIp = 2*ip;
         tr      = a[indexIp];
         ti      = a[indexIp+1];
         a[indexIp  ] = a[indexI  ] - tr;
         a[indexIp+1] = a[indexI+1] - ti;
         a[indexI  ]  = a[indexI  ] + tr;
         a[indexI+1]  = a[indexI+1] + ti;
        }

     for (j=1; j<le1; j++)
        {for (i=j; i<n; i=i+le)
            {ip      = i + le1;
             indexI  = 2*i;
             indexIp = 2*ip;
             tr      = a[indexIp  ]*ur - a[indexIp+1]*ui;
             ti      = a[indexIp+1]*ur + a[indexIp  ]*ui;
             a[indexIp  ] = a[indexI  ] - tr;
             a[indexIp+1] = a[indexI+1] - ti;
             a[indexI  ]  = a[indexI  ] + tr;
             a[indexI+1]  = a[indexI+1] + ti;
            }
 
         ur0 = ur;
         ur  = ur*wr - ui*wi;
         ui  = ui*wr + ur0*wi;
        }
    }
 
 for (i=1; i<n/2; i++)
    {temp         = a[2*i];
     a[2*i]       = a[2*(n-i)];
     a[2*(n-i)]   = temp;
     temp         = a[2*i+1];
     a[2*i+1]     = a[2*(n-i)+1];
     a[2*(n-i)+1] = temp;
    }

 return;
}
