/* -*- mode: c++; c-basic-offset: 4; -*- */
#include <time.h>
#include <iostream>
#include <stdio.h>
#include "NameClient.hh"
#include "lmsg/ErrorList.hh"
#include <cstdlib>

using namespace lmsg;
using namespace std;

int 
main(int argc, const char** argv) {

    //----------------------------------  Parse the arguments
    bool syntax(false);
    int ident(-1), debug(0);
    for (int i=1 ; i<argc-1 ; i++) {
        string argi(argv[i]);
        if (argi == "-id") {
	    ident = strtol(argv[++i], 0, 0);
        } else if (argi == "-debug") {
	    debug = strtol(argv[++i], 0, 0);
	} else {
	    syntax = true;
	}
    }

    if (syntax || argc < 2) {
        cerr << "Error in command line. Command syntax: " << endl;
	cerr << "    " << argv[0] << " [-debug <debug-level>] "
	     << "{-id <server-id> | <server-name>}" << endl;
	return 1;
    }

    //----------------------------------  Get the name server entry
    NameClient nc;
    nc.setDebug(debug);
    NameData   Name;
    error_type rc(OK);
    if (ident < 0) rc = nc.lookup(argv[argc-1], Name);
    else           rc = nc.lookupID(ident, Name);
    
    if (rc) {
        switch(rc) {
	case BadAddress:
	case TimeOut:
	    cerr << "Unable to reach name server at: " << nc.getDomain() 
		 << endl;
	    break;
	case Invalid:
	case Failure:
	    if (ident < 0) cerr << "Server name: " << argv[argc-1] 
				<< " is not registered." << endl;
	    else           cerr << "Server ID: " << ident
				<< " does not exist." << endl;
	    break;
	case SystemError:
	    perror("System Error");
	    break;
	default:
	    cerr << "Unexpected error" << endl;
	}
	
        return rc;
    }

    //----------------------------------  Print the result
    cout << "Address entry for " << argv[argc-1] << ":" << endl;
    cout << "Name:    " << Name.getName() << endl;
    cout << "Type:    " << Name.getPType() << endl;
    cout << "Address: " << Name.getAddr() << endl;
    return 0;
}
